from ssoss.dynamic_road_object import *
from ssoss.motion_road_object import *
from ssoss.process_road_objects import *
from ssoss.process_video import *
from ssoss.static_road_object import *
import importlib.metadata

__version__ = importlib.metadata.version("ssoss")

